#include "die.h"
#include <netdb.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

static void handleConnection(int clientSock, int listenSock);
static void handleRequest(FILE *rx, FILE *tx);

// Initialisiert den Server und nimmt Anfragen an
static void initServer(short unsigned int port) {
    //ignore SIGPIPE
    struct sigaction sa;
    sa.sa_handler = SIG_IGN;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;

    if( sigaction(SIGPIPE, &sa, NULL) != 0 ) die("Could not set sigaction");

    //init accept
    int listenSock = socket(AF_INET6, SOCK_STREAM, 0);
    if ( listenSock == -1 ) die("Fehler beim Aufbau des Servers");
    struct sockaddr_in6 addr = {
        .sin6_family = AF_INET6,
        .sin6_port   = htons(port),
        .sin6_addr   = in6addr_any,
    };
    if ( bind(listenSock, (struct sockaddr *) &addr, sizeof(addr)) != 0 ) die("Fehler beim Aufbau des Servers");
    if ( listen(listenSock, SOMAXCONN) != 0 ) die("Fehler beim Aufbau des Servers");

    while ( 1 ) {
        int clientSock = accept(listenSock, NULL, NULL);
        if ( clientSock == -1 ) { perror("Warnung: Client konnte nicht akzeptiert werden."); continue; }
        handleConnection(clientSock, listenSock);
    }
}



static void handleConnection(int clientSock, int listenSock) {
    //if pthread just bbput

    //fork()
    pid_t pid;
    if( (pid = fork()) == -1 ) die("fork");
    if(pid != 0) {//parent process
        close(clientSock);//only in parent process
        return;
    }
    if( close(listenSock) ) perror("close listenSock");

    //open file streams
    FILE *rx, *tx;
    int clientSock_Copy = dup(clientSock);
    if(clientSock_Copy < 0) die("dup");

    if( (rx = fdopen(clientSock, "r")) == NULL) die("fdopen");
    if( (tx = fdopen(clientSock_Copy, "w")) == NULL) die("fdopen");

    //handleRequest()
    handleRequest(rx, tx);

    //close connection
    if( fclose(tx) ) die("fclose tx");
    if( fclose(rx) ) die("fclose rx");
    
    //exit
    exit(EXIT_SUCCESS);//child exits
}

static void handleRequest(FILE *rx, FILE *tx) {
    rx = (FILE *) rx;
    fprintf(tx, "hi\n");
}

int main() {
    initServer(2020);
}
