#include <stdatomic.h>
#include "jbuffer.h"
#include "sem.h"

struct BNDBUF
{
    int *buff;
    int size;
    int insertpos; // next insert pos
    SEM *freeSpaceSem; // count of free slots
    int extractpos; // next extract pos
    SEM *usedSpaceSem; // count of used slots
};

BNDBUF *bbCreate(size_t size) {
    if(size > __INT_MAX__) return NULL;//to big

    BNDBUF *bb = calloc(1, sizeof(BNDBUF));//initialize everithing with NULL
    if(bb == NULL) return NULL;//malloc failed

    bb->buff = malloc(size * sizeof(int));
    if(bb->buff == NULL) {//malloc failed
        bbDestroy(bb);
        return NULL;
    }
    bb->size = (int) size;

    //init SEMs
    bb->freeSpaceSem = semCreate((int) size);
    if(bb->freeSpaceSem == NULL) {//semCreate failed
        bbDestroy(bb);
        return NULL;
    }

    bb->usedSpaceSem = semCreate(0);
    if(bb->usedSpaceSem == NULL) {//semCreate failed
        bbDestroy(bb);
        return NULL;
    }

    bb->insertpos = 0;
    bb->extractpos = 0;
    return bb;
}

void bbDestroy(BNDBUF *bb) {
    if(bb == NULL) return;
    free(bb->buff);//each does nothing if NULL
    semDestroy(bb->freeSpaceSem);
    semDestroy(bb->usedSpaceSem);
    free(bb);
}

void bbPut(BNDBUF *bb, int value) {
    P(bb->freeSpaceSem);
    bb->buff[bb->insertpos] = value;
    bb->insertpos++;
    if(bb->insertpos >= bb->size) bb->insertpos = 0;
    V(bb->usedSpaceSem);
}

int bbGet(BNDBUF *bb) {
    int value, extractpos, newExtractpos;
    P(bb->usedSpaceSem);
    do{
        extractpos = bb->extractpos;
        value = bb->buff[extractpos];
        newExtractpos = extractpos + 1;
        if(newExtractpos >= bb->size) newExtractpos = 0;
    } while(! atomic_compare_exchange_strong(&(bb->extractpos), &extractpos, newExtractpos));
    V(bb->freeSpaceSem);
    return value;
}
