#!/usr/bin/env node
const fs = require('fs');
const markedpp = require('markedpp');
const markdownit = require('markdown-it')();

markdownit.use(require('markdown-it-highlightjs'));

let name = process.argv[1];
let outfile;
let infile;
let preprocessorOutfile;

//parse params
for(let i=2; i < process.argv.length; i++) {
    switch (process.argv[i]) {
        case '-o':
            i++
            if(i >= process.argv.length) {
                console.error("-o expected out file");
                return -1;
            }
            outfile = process.argv[i];
            break;
        
        case '-m':
            i++
            if(i >= process.argv.length) {
                console.error("-m expected preprocessed md out file");
                return -1;
            }
            preprocessorOutfile = process.argv[i];
            break;

        case '-h':
            console.log("usage: " + name + " -o <outfile> [-m <preprocessor outfile>] <infile>");
            break;
    
        default:
            if(i == process.argv.length - 1) {
                infile = process.argv[i];
            } else {
                console.error("unknown argument: " + process.argv[i]);
            }
            break;
    }
}

//check required params
if(!outfile) {
    console.error("missing param outfile")
    return -1;
}
if(!infile) {
    console.error("missing param infile")
    return -1;
}


fs.readFile(infile, {encoding: 'utf-8'}, (err, data) => {
    //markdown preprocessor
    markedpp(data, (err, result) => {
        if(err) {
            console.log(err);
        } else {
            if(preprocessorOutfile) {
                fs.writeFileSync(preprocessorOutfile, result);
            }
            //markdown-it
            let html = '<link rel="stylesheet" href="node_modules/highlight.js/styles/default.css">';
            html += markdownit.render(result);

            fs.writeFileSync(outfile, html);
        }
    });
});


