#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include "sem.h"

struct SEM
{
    pthread_mutex_t m;
    pthread_cond_t c;
    int value;
};

SEM *semCreate(int initVal) {
    SEM *sem = malloc(sizeof(SEM));
    if(sem == NULL) return NULL;// malloc failed
    int err=0;
    err = pthread_mutex_init(&sem->m, NULL);
    if(err) {
        free(sem);
        errno = err;
        return NULL;
    }
    err = pthread_cond_init(&sem->c, NULL);
    if (err) {
        pthread_mutex_destroy(&sem->m);
        free(sem);
        errno = err;
        return NULL;
    }

    sem->value = initVal;
    return sem;
}

void semDestroy(SEM *sem) {
    if(sem == NULL) return;
    
    pthread_cond_destroy(&sem->c);//ignore returned errors
    pthread_mutex_destroy(&sem->m);//ignore returned errors
    free(sem);
}

void P(SEM *sem) {//fast mutex has only EINVAL error
    pthread_mutex_lock(&sem->m);
    while(sem->value <= 0) {
        pthread_cond_wait(&sem->c, &sem->m);
    }
    sem->value--;

    pthread_mutex_unlock(&sem->m);
}

void V(SEM *sem) {//fast mutex has only EINVAL error
    pthread_mutex_lock(&sem->m);
    sem->value++;
    pthread_cond_signal(&sem->c);

    pthread_mutex_unlock(&sem->m);
}
