#include "die.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

static void sigchldHandler(int sig);
static int pcount = 0;

int main() {
    struct sigaction handleSIGCHLD =
        {
        .sa_handler = sigchldHandler,
        .sa_flags   = SA_RESTART,
        };
    if( sigaction(SIGCHLD, &handleSIGCHLD, NULL) != 0 ) die("Could not set sigaction");

    //mask sigchld
    sigset_t newMask;
    sigset_t oldMask;
    sigemptyset(&newMask);
    sigaddset(&newMask, SIGCHLD);
    if (sigprocmask(SIG_BLOCK, &newMask, &oldMask) == -1) die("sigprocmask");

    //do critical stuff
    pid_t pid = fork();
    if(pid < 0) die("fork");
    if(pid == 0) {
        sleep(5);
        exit(EXIT_SUCCESS);
    }
    pcount++;

    //alte signal mask wiederherstellen
    if (sigprocmask(SIG_SETMASK, &oldMask, NULL) == -1) die("sigprocmask");


    //passives warten
    if (sigprocmask(SIG_BLOCK, &newMask, &oldMask) == -1) die("sigprocmask");
    while (pcount > 0) {
        sigsuspend(&oldMask);
    }
}

static void sigchldHandler(int sig) {
    // Collect zombie processes
    (void)sig; //unused param
    pid_t pid;
    int event;
    while ((pid = waitpid(-1, &event, WNOHANG)) > 0) {
        // collect all zombies (prevent zombies if signal got lost because of masking)
        pcount--;
    }
}
